package com.google;

import com.amazonaws.services.textract.model.DetectDocumentTextResult;
import com.google.api.gax.longrunning.OperationFuture;
import com.google.cloud.storage.*;
import com.google.cloud.vision.v1.*;
import com.google.cloud.vision.v1.Word;
import com.google.protobuf.ByteString;
import com.google.protobuf.Descriptors;
import com.google.protobuf.util.JsonFormat;
import org.apache.poi.xwpf.usermodel.XWPFDocument;
import org.apache.poi.xwpf.usermodel.XWPFParagraph;
import org.apache.poi.xwpf.usermodel.XWPFRun;
import org.apache.poi.xwpf.usermodel.XWPFTable;
import org.json.JSONException;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static com.google.Readability.calculateMetrics;
import static com.google.Readability.getEducationLevel;

public class DocumentProcess {

    private String inputPath = "D:\\Temp\\Input";
    private String outputPath = "D:\\Temp\\Output\\";
    private String name;
    private List<DetectDocumentTextResult> result;
    LinkedHashMap<String, List<AnnotateImageResponse>> fileAndResponses = new LinkedHashMap<>();
    List<String> filenames = new ArrayList<>();

    //Google
    String bucketName = "service_bucket_ocr";
    String gcsSourcePath = "gs://service_bucket_ocr/inputs/";
    String gcsDestinationPath = "gs://service_bucket_ocr/outputs/";

    public DocumentProcess() {
    }

    public void processDocumentGoogle(String name) throws IOException {
        this.name = name;

        File file = new File(inputPath);
        System.out.println("Processing documents: " + file.listFiles().length);
        for (int f = 0; f < file.listFiles().length; f++) {
            System.out.println(file.getPath() + "\\" + file.listFiles()[f].getName());
            System.out.println("Absolut path in google: " + file.listFiles()[f].getAbsolutePath());

            String filename = file.listFiles()[f].getName();

            if (filename.toLowerCase().endsWith(".pdf")) {
                processDocumentPDFGoogle(file.listFiles()[f]);
            } else {
                List<AnnotateImageRequest> requests = new ArrayList<>();

                ByteString imgBytes = ByteString.readFrom(new FileInputStream(file.listFiles()[f].getAbsolutePath()));

                Image img = Image.newBuilder().setContent(imgBytes).build();
                Feature feat = Feature.newBuilder().setType(Feature.Type.TEXT_DETECTION).build();
                AnnotateImageRequest request =
                        AnnotateImageRequest.newBuilder().addFeatures(feat).setImage(img).build();
                requests.add(request);

                // Initialize client that will be used to send requests. This client only needs to be created
                // once, and can be reused for multiple requests. After completing all of your requests, call
                // the "close" method on the client to safely clean up any remaining background resources.
                List<AnnotateImageResponse> responses = new ArrayList<>();
                try (ImageAnnotatorClient client = ImageAnnotatorClient.create()) {
                    BatchAnnotateImagesResponse response = client.batchAnnotateImages(requests);
                    //System.out.println(response);

                    responses.addAll(response.getResponsesList());
                }
                filenames.add(filename);
                System.out.println(filename + " and " + responses.size());
                fileAndResponses.put(filename, responses);
            }
        }
        if(fileAndResponses.size() > 0){
            createOutput();
        }
    }

    public void processDocumentPDFGoogle(File file) throws IOException {

        String filename = file.getName();
        String sourcePath = gcsSourcePath;
        String destinationPath = gcsDestinationPath + name + "_";

        // Initialize client that will be used to send requests. This client only needs to be created
        // once, and can be reused for multiple requests. After completing all of your requests, call
        // the "close" method on the client to safely clean up any remaining background resources.
        try (ImageAnnotatorClient client = ImageAnnotatorClient.create()) {
            List<AsyncAnnotateFileRequest> requests = new ArrayList<>();

            //import document
            Storage storage = StorageOptions.getDefaultInstance().getService();
            BlobId blobId = BlobId.of(bucketName, "inputs/" + filename);
            BlobInfo blobInfo = BlobInfo.newBuilder(blobId).build();
            storage.create(blobInfo, Files.readAllBytes(Paths.get(file.getAbsolutePath())));

            // Set the GCS source path for the remote file.
            GcsSource gcsSource = GcsSource.newBuilder().setUri(sourcePath + filename).build();

            // Create the configuration with the specified MIME (Multipurpose Internet Mail Extensions)
            // types
            InputConfig inputConfig =
                    InputConfig.newBuilder()
                            .setMimeType(
                                    "application/pdf") // Supported MimeTypes: "application/pdf", "image/tiff"
                            .setGcsSource(gcsSource)
                            .build();

            // Set the GCS destination path for where to save the results.
            GcsDestination gcsDestination =
                    GcsDestination.newBuilder().setUri(destinationPath).build();

            // Create the configuration for the System.output with the batch size.
            // The batch size sets how many pages should be grouped into each json System.output file.
            OutputConfig outputConfig =
                    OutputConfig.newBuilder().setBatchSize(2).setGcsDestination(gcsDestination).build();

            // Select the Feature required by the vision API
            Feature feature = Feature.newBuilder().setType(Feature.Type.DOCUMENT_TEXT_DETECTION).build();

            // Build the OCR request
            AsyncAnnotateFileRequest request =
                    AsyncAnnotateFileRequest.newBuilder()
                            .addFeatures(feature)
                            .setInputConfig(inputConfig)
                            .setOutputConfig(outputConfig)
                            .build();

            requests.add(request);

            // Perform the OCR request
            OperationFuture<AsyncBatchAnnotateFilesResponse, OperationMetadata> response =
                    client.asyncBatchAnnotateFilesAsync(requests);

            System.out.println("Waiting for the operation to finish.");

            // Wait for the request to finish. (The result is not used, since the API saves the result to
            // the specified location on GCS.)
            List<AsyncAnnotateFileResponse> result =
                    response.get(180, TimeUnit.SECONDS).getResponsesList();

            // Get the destination location from the gcsDestinationPath
            Pattern pattern = Pattern.compile("gs://([^/]+)/(.+)");
            Matcher matcher = pattern.matcher(destinationPath);

            if (matcher.find()) {
                String bucketName2 = matcher.group(1);
                String prefix = matcher.group(2);

                // Get the list of objects with the given prefix from the GCS bucket
                Bucket bucket = storage.get(bucketName2);
                com.google.api.gax.paging.Page<Blob> pageList = bucket.list(Storage.BlobListOption.prefix(prefix));
                List<Blob> firstOutputFile = new ArrayList<>();

                // List objects with the given prefix.
                System.out.println("Output files:");
                for (Blob blob : pageList.iterateAll()) {
                    System.out.println(blob.getName());
                    // Process the first System.output file from GCS.
                    // Since we specified batch size = 2, the first response contains
                    // the first two pages of the input file.
                    firstOutputFile.add(blob);
                }

                List<AnnotateImageResponse> responses = new ArrayList<>();
                //Clean output folder in Google Storage
                for(Blob output : firstOutputFile) {
                    // Get the contents of the file and convert the JSON contents to an AnnotateFileResponse
                    // object. If the Blob is small read all its content in one request
                    // (Note: the file is a .json file)
                    // Storage guide: https://cloud.google.com/storage/docs/downloading-objects
                    String jsonContents = new String(output.getContent());
                    AnnotateFileResponse.Builder builder = AnnotateFileResponse.newBuilder();
                    JsonFormat.parser().merge(jsonContents, builder);

                    // Build the AnnotateFileResponse object
                    AnnotateFileResponse annotateFileResponse = builder.build();

                    // Parse through the object to get the actual response for the first page of the input file.
                    //List<AnnotateImageResponse> responses = annotateFileResponse.getResponsesList();
                    responses.addAll(annotateFileResponse.getResponsesList());

                    //Delete output file in Google Storage
                    BlobId o = BlobId.of(bucketName2, output.getName());
                    if (storage.delete(o)) {
                        System.out.println(output.getName() + " successfully deleted");
                    }
                }

                filenames.add(filename);
                fileAndResponses.put(filename, responses);

                //Clean input folder in Google Storage
                BlobId i = BlobId.of(bucketName2, "inputs/" + file.getName());
                if (storage.delete(i)) {
                    System.out.println("inputs/" + file.getName() + " successfully deleted");
                }

            } else {
                System.out.println("No MATCH");
            }
        } catch (InterruptedException e) {
            e.printStackTrace();
        } catch (ExecutionException e) {
            e.printStackTrace();
        } catch (TimeoutException e) {
            e.printStackTrace();
        }
    }

    private void createOutput() throws IOException {
        //for json
        //ArrayList<Block> blocksInFile = new ArrayList<>();
        for(String filename : filenames) {
            //for docx
            ArrayList<Block> blocksInFile = new ArrayList<>();
            int pageCounter = 0;
            List<AnnotateImageResponse> responses = fileAndResponses.get(filename);
            StringBuilder textToCalculate = new StringBuilder();
            for (AnnotateImageResponse res : responses) {
                for (Page page : res.getFullTextAnnotation().getPagesList()) {
                    pageCounter++;
                    int paragraphCounterOnPage = 0;
                    for (com.google.cloud.vision.v1.Block block : page.getBlocksList()) {
                        String text = "";
                        paragraphCounterOnPage++;
                        for (Paragraph paragraph : block.getParagraphsList()) {
                            StringBuilder newParagraph = new StringBuilder();
                            for (Word word : paragraph.getWordsList()) {
                                if (!newParagraph.equals(""))
                                    newParagraph.append(" ");

                                for (Symbol symbol : word.getSymbolsList()) {
                                    newParagraph = newParagraph.append(symbol.getText());
                                }
                            }



                            textToCalculate.append(" " + newParagraph.toString());
                            if(text.isEmpty())
                                text = newParagraph.toString();
                            else
                                text = text + newParagraph.toString();

                        }
                        text = text.replaceAll("\"", "\\\\\"");

                        String last = "false";
                        if (paragraphCounterOnPage == page.getBlocksList().size())
                            last = "true";
                        blocksInFile.add(
                                new Block(text, text, pageCounter, paragraphCounterOnPage, page.getBlocksList().size(), filename, last, "Text_block")
                        );
                    }
                }
            }
            //for docx
            if(blocksInFile.size() > 0) {
                List kincaidResults = calculateMetrics(textToCalculate.toString());
                String educationLevel = getEducationLevel(kincaidResults);

                writeToDocxFile(blocksInFile, educationLevel);
            }

        }
        //for JSON
        /*if (blocksInFile.size() > 0) {
            storeJsonStructure(blocksInFile);
        }*/

    }

    static class Block {
        String Text;
        String FinalText;
        Integer PageNumber;
        Integer BlockOnPageNumber;
        Integer BlocksOnPageCount;
        String InitialInputFileName;
        String Last;
        String BlockType;

        public Block(String text, String finalText, Integer pageNumber, Integer blockOnPageNumber, Integer blocksOnPageCount, String initialInputFileName, String last, String blockType) {
            Text = text;
            FinalText = finalText;
            PageNumber = pageNumber;
            BlockOnPageNumber = blockOnPageNumber;
            BlocksOnPageCount = blocksOnPageCount;
            InitialInputFileName = initialInputFileName;
            Last = last;
            BlockType = blockType;
        }

        public String getValues(){
            return "{ \"Text\": \"" + Text + "\",\"FinalText\": \"" + FinalText + "\",\"PageNumber\": " + PageNumber + ',' +
                    "\"BlockOnPageNumber\": " + BlockOnPageNumber + ",\"BlocksOnPageCount\": " + BlocksOnPageCount + ',' +
                    "\"InitialInputFileName\": \"" + InitialInputFileName + "\",\"Last\": \"" + Last + "\", \"BlockType\": \"" + BlockType + "\"}";
        }
    }

    private void storeJsonStructure(ArrayList<Block> arrayFinal) throws IOException {
        System.out.println("Creating JSON");

        BufferedWriter output = new BufferedWriter(new OutputStreamWriter(
                new FileOutputStream(outputPath + name + ".json"),
                StandardCharsets.UTF_8
        ));

        output.write("{\"document\": [");

        // write under header
        output.write("{\"blocks\": [");

        boolean firstBlock = true;
        for(Block block : arrayFinal) {

            String jsonResponse = block.getValues();

            if (firstBlock) {
                firstBlock = false;
            } else {
                output.write(",");
            }

            output.write(jsonResponse);
        }
        output.write("]}]}");
        output.close();
    }

    private XWPFDocument createNewDocxDocument(){
        //Blank Document
        XWPFDocument document = new XWPFDocument();
        return document;
    }

    private void writeToDocxFile(List<Block> arrayFinal, String educationLevel) throws IOException{


        XWPFDocument document = createNewDocxDocument();
        //Write the Document in file system
        File file = new File(outputPath + "\\" + name + "\\" + educationLevel + "\\"+ arrayFinal.get(0).InitialInputFileName + ".docx");
        file.getParentFile().mkdirs();
        FileOutputStream out= new FileOutputStream(outputPath + "\\" + name + "\\" + educationLevel + "\\" + arrayFinal.get(0).InitialInputFileName + ".docx");;

        for(int i = 0; i < arrayFinal.size(); i++) {
            String rakeName = BlockNames.newNameFuction(arrayFinal.get(i).Text);
            String coreNLPName = BlockNames.newNameFuctionST(arrayFinal.get(i).Text);

            //create Paragraph
            XWPFParagraph paragraph = document.createParagraph();
            XWPFRun run = paragraph.createRun();
            if(i != 0)
                run.addBreak();
            run.setBold(true);
            run.setText(rakeName + " (" + coreNLPName + ")", 0);

            XWPFTable table = document.createTable(1,1);
            table.setWidth("100%");
            XWPFParagraph paragraph2 = table.getRow(0).getCell(0).getParagraphs().get(0);
            paragraph2.setIndentFromLeft(100);
            paragraph2.setIndentFromRight(100);
            paragraph2.createRun().setText(arrayFinal.get(i).Text);

            //Close document
            System.out.println(outputPath + "\\" + educationLevel + "\\" + arrayFinal.get(i).InitialInputFileName + ".docx" + " written successfully");
        }
        document.write(out);
        out.close();
    }

    public void deleteFiles(){
        File file = new File(inputPath);
        int counter = file.list().length;
        while(counter > 0) {
            System.out.println("Deleting " + file.getPath() + "\\" + file.list()[0]);
            new File(file.getPath() + "\\" + file.list()[0]).delete();
            counter--;
        }

        if(file.list().length > 0){
            System.out.println("NOT ALL FILES WERE DELETED IN INPUT FOLDER!");
        }
    }
}
